/*
 * Copyright 2013 Canonical Ltd.
 *
 * Authors:
 *   Charles Kerr <charles.kerr@canonical.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libical/ical.h>
#include <libical/icaltime.h>
#include <libecal/libecal.h>
#include <libedataserver/libedataserver.h>

#include "planner-eds.h"

struct _IndicatorDatetimePlannerEdsPrivate
{
  GSList * sources;
  GSList * views;
  GCancellable * cancellable;
  ESourceRegistry * source_registry;
};

typedef IndicatorDatetimePlannerEdsPrivate priv_t;

G_DEFINE_TYPE_WITH_PRIVATE (IndicatorDatetimePlannerEds,
                            indicator_datetime_planner_eds,
                            INDICATOR_TYPE_DATETIME_PLANNER)

G_DEFINE_QUARK ("source-client", source_client)
G_DEFINE_QUARK ("source-view", source_view)

/***
****
****  my_get_appointments() helpers
****
***/

struct get_appointments_task_data
{
  /* how many subtasks are still running on */
  int subtask_count;

  /* the list of appointments to be returned */
  GSList * appointments;

  /* ensure that recurring events don't get multiple IndicatorDatetimeAppts */
  GHashTable * added;
};

static void
get_appointments_task_data_free (gpointer gdata)
{
  struct get_appointments_task_data * data = gdata;
  g_hash_table_destroy (data->added);
  g_slice_free (struct get_appointments_task_data, data);
}

static void
on_all_subtasks_done (GTask * task)
{
  struct get_appointments_task_data * data = g_task_get_task_data (task);
  g_task_return_pointer (task, data->appointments, NULL);
  g_object_unref (task);
}

struct get_appointments_subtask_data
{
  /* The parent task */
  GTask * task;

  /* The appointment's color coding. The subtask owns this string */
  gchar * color;

  /* The uid of a appointment's source. he subtask owns this string */
  gchar * source_uid;

  /* The default timezone for time conversion */
  ICalTimezone *default_timezone;
};

static void
on_subtask_done (gpointer gsubdata)
{
  struct get_appointments_subtask_data * subdata;
  GTask * task;
  struct get_appointments_task_data * data;

  subdata = gsubdata;
  task = subdata->task;

  /* free the subtask data */
  g_free (subdata->color);
  g_free (subdata->source_uid);
  g_slice_free (struct get_appointments_subtask_data, subdata);

  /* poke the task */
  data = g_task_get_task_data (task);
  if (g_atomic_int_dec_and_test (&data->subtask_count))
    on_all_subtasks_done (task);
}

static time_t
my_icaltime_to_timet (ICalTime *icaltime,
                      ICalTimezone *default_timezone)
{
  ICalTime *tt;
  time_t res;

  tt = i_cal_time_convert_to_zone (icaltime, default_timezone);
  res = i_cal_time_as_timet_with_zone (tt, default_timezone);
  g_clear_object (&tt);

  return res;
}

static gboolean
my_get_appointments_foreach (ICalComponent *icomp,
                             ICalTime *instance_start,
                             ICalTime *instance_end,
                             gpointer gsubdata,
                             G_GNUC_UNUSED GCancellable *cancellable,
                             G_GNUC_UNUSED GError **error)
{
  ECalComponent *component = e_cal_component_new_from_icalcomponent (i_cal_component_clone (icomp));
  const ECalComponentVType vtype = e_cal_component_get_vtype (component);
  struct get_appointments_subtask_data * subdata = gsubdata;
  struct get_appointments_task_data * data = g_task_get_task_data (subdata->task);

  if ((vtype == E_CAL_COMPONENT_EVENT) || (vtype == E_CAL_COMPONENT_TODO))
    {
      const gchar * uid = NULL;
      ICalPropertyStatus status = 0;

      uid = e_cal_component_get_uid (component);
      status = e_cal_component_get_status (component);

      if ((uid != NULL) &&
          (!g_hash_table_contains (data->added, uid)) &&
          (status != I_CAL_STATUS_COMPLETED) &&
          (status != I_CAL_STATUS_CANCELLED))
        {
          GSList * alarm_uids;
          GSList * l;
          GSList * recur_list;
          ECalComponentId *id;
          ECalComponentText *text;
          struct IndicatorDatetimeAppt * appt;

          appt = g_slice_new0 (struct IndicatorDatetimeAppt);

          /* Determine whether this is a recurring event.
             NB: icalrecurrencetype supports complex recurrence patterns;
             however, since design only allows daily recurrence,
             that's all we support here. */
          recur_list = e_cal_component_get_rrules (component);
          for (l=recur_list; l!=NULL; l=l->next)
            {
              ICalRecurrence * recur = l->data;
              appt->is_daily |= ((i_cal_recurrence_get_freq (recur) == I_CAL_DAILY_RECURRENCE)
                                  && (i_cal_recurrence_get_interval (recur) == 1));
            }
          g_slist_free_full (recur_list, g_object_unref);
          
          /* Get rid, this will be used in openning
             calendar appointments in gnome-calendar */
          id = e_cal_component_get_id (component);
          if (e_cal_component_id_get_rid (id))
            appt->rid = g_strdup (e_cal_component_id_get_rid (id));
          e_cal_component_id_free (id);

          /* Get summary. We are using empty string,
             evolution uses "No_Summary" */
          text = e_cal_component_get_summary (component);
          appt->summary = g_strdup (text && e_cal_component_text_get_value (text) ? e_cal_component_text_get_value (text) : "");

          /* Get appointment uid */ 
          appt->uid = g_strdup (uid);

          /* get appointment begin & end */
          appt->begin = g_date_time_new_from_unix_local (my_icaltime_to_timet (instance_start, subdata->default_timezone));
          appt->end = g_date_time_new_from_unix_local (my_icaltime_to_timet (instance_end, subdata->default_timezone));

          /* Get source color */
          appt->color = g_strdup (subdata->color);

          /* Get source uid */
          appt->source_uid = g_strdup (subdata->source_uid);

          /* Get appointment.type */
          appt->is_event = vtype == E_CAL_COMPONENT_EVENT;
          
          /* Appointment has alarm ? */
          alarm_uids = e_cal_component_get_alarm_uids (component);
          appt->has_alarms = alarm_uids != NULL;
          g_slist_free_full (alarm_uids, g_free);

          data->appointments = g_slist_prepend (data->appointments, appt);
          g_hash_table_add (data->added, g_strdup(uid));
        }
    }

  g_clear_object (&component);

  return G_SOURCE_CONTINUE;
}

/***
****  IndicatorDatetimePlanner virtual funcs
***/

static void
my_get_appointments (IndicatorDatetimePlanner  * planner,
                     GDateTime                 * begin_datetime,
                     GDateTime                 * end_datetime,
                     GAsyncReadyCallback         callback,
                     gpointer                    user_data)
{
  GSList * l;
  priv_t * p;
  const char * str;
  ICalTimezone * default_timezone;
  struct get_appointments_task_data * data;
  const int64_t begin = g_date_time_to_unix (begin_datetime);
  const int64_t end = g_date_time_to_unix (end_datetime);
  GTask * task;
  gboolean subtasks_added;

  p = INDICATOR_DATETIME_PLANNER_EDS (planner)->priv;

  /**
  ***  init the default timezone
  **/

  default_timezone = NULL;

  if ((str = indicator_datetime_planner_get_timezone (planner)))
    {
      default_timezone = i_cal_timezone_get_builtin_timezone (str);

      if (default_timezone == NULL) /* maybe str is a tzid? */
        default_timezone = i_cal_timezone_get_builtin_timezone_from_tzid (str);
    }

  /**
  ***  walk through the sources to build the appointment list
  **/

  data = g_slice_new0 (struct get_appointments_task_data);
  data->added = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  task = g_task_new (planner, p->cancellable, callback, user_data);
  g_task_set_task_data (task, data, get_appointments_task_data_free);

  subtasks_added = FALSE;
  for (l=p->sources; l!=NULL; l=l->next)
    {
      ESource * source;
      ECalClient * client;
      struct get_appointments_subtask_data * subdata;

      source = l->data;
      client = g_object_get_qdata (l->data, source_client_quark());
      if (client == NULL)
        continue;

      if (default_timezone != NULL)
        e_cal_client_set_default_timezone (client, default_timezone);

      subdata = g_slice_new (struct get_appointments_subtask_data);
      subdata->task = task;
      subdata->source_uid = g_strdup (e_source_get_uid (source));
      subdata->color = e_source_selectable_dup_color (e_source_get_extension (source, E_SOURCE_EXTENSION_CALENDAR));
      subdata->default_timezone = default_timezone;

      g_atomic_int_inc (&data->subtask_count);
      subtasks_added = TRUE;
      e_cal_client_generate_instances (client,
                                       begin,
                                       end,
                                       p->cancellable,
                                       my_get_appointments_foreach,
                                       subdata,
                                       on_subtask_done);
    }

  if (!subtasks_added)
    on_all_subtasks_done (task);
}

static GSList *
my_get_appointments_finish (IndicatorDatetimePlanner  * self  G_GNUC_UNUSED,
                            GAsyncResult              * res,
                            GError                   ** error)
{
  return g_task_propagate_pointer (G_TASK(res), error);
}

static gboolean
planner_has_app (const gchar * app_name)
{
  gchar * app;

  app = g_find_program_in_path (app_name);
  if (app == NULL)
    return FALSE;

  g_free (app);
  return TRUE;
}

static gboolean
my_is_configured (IndicatorDatetimePlanner * planner)
{
  IndicatorDatetimePlannerEds * self;

  if (planner_has_app ("gnome-calendar") || planner_has_app ("gnome-calendar"))
    {
      /* see if there are any calendar sources */
      self = INDICATOR_DATETIME_PLANNER_EDS (planner);
      return self->priv->sources != NULL;
    }
  else 
    {
      return FALSE;
    }
}

static void
my_activate (IndicatorDatetimePlanner * self G_GNUC_UNUSED)
{
  gchar * command;
  GError * err = NULL;

  if (planner_has_app ("gnome-calendar"))
    {
      command = g_strdup ("gnome-calendar");
      if (!g_spawn_command_line_async (command, &err))
        {
          g_warning ("Unable to start %s: %s", command, err->message);
          g_error_free (err);
        }

      g_free (command);
    }
  else if (planner_has_app ("evolution"))
    {
      command = g_strdup ("evolution -c calendar");
      if (!g_spawn_command_line_async (command, &err))
        {
          g_warning ("Unable to start %s: %s", command, err->message);
          g_error_free (err);
        }

      g_free (command);
    }
}

static void
my_activate_time (IndicatorDatetimePlanner * self G_GNUC_UNUSED,
                  GDateTime * activate_time)
{
  gchar * isodate;
  gchar * command;
  GError * err = NULL;

  if (planner_has_app ("gnome-calendar"))
    {
      isodate = g_date_time_format (activate_time, "%m/%d/%Y");
      command = g_strdup_printf ("gnome-calendar --date %s", isodate);
      if (!g_spawn_command_line_async (command, &err))
        {
          g_warning ("Unable to start %s: %s", command, err->message);
          g_error_free (err);
        }

      g_free (command);
      g_free (isodate);
    }
  else if (planner_has_app ("evolution"))
    {
      isodate = g_date_time_format (activate_time, "%Y%m%d");
      command = g_strdup_printf ("evolution \"calendar:///?startdate=%s\"", isodate);
      if (!g_spawn_command_line_async (command, &err))
        {
          g_warning ("Unable to start %s: %s", command, err->message);
          g_error_free (err);
        }

      g_free (command);
      g_free (isodate);
    }
}

static void
my_activate_appointment (IndicatorDatetimePlanner * planner G_GNUC_UNUSED,
                         const gchar * uuid)
{
  gchar * command;
  GError * error = NULL;
  
  command = g_strdup_printf ("gnome-calendar -u \"%s\"", uuid);
  if (!g_spawn_command_line_async (command, &error))
    {
      g_warning ("Unable to start %s: %s", command, error->message);
      g_error_free (error);
    }
  g_free (command);
}

/***
****  Source / Client Wrangling
***/
static void
on_view_objects_added (ECalClientView * view    G_GNUC_UNUSED,
                       EClient        * client  G_GNUC_UNUSED,
                       gpointer         gself)
{
  indicator_datetime_planner_emit_appointments_changed (gself);
}

static void
on_view_objects_modified (ECalClientView * view    G_GNUC_UNUSED,
                          EClient        * client  G_GNUC_UNUSED,
                          gpointer         gself)
{
  indicator_datetime_planner_emit_appointments_changed (gself);
}

static void
on_view_objects_removed (ECalClientView * view    G_GNUC_UNUSED,
                         EClient        * client  G_GNUC_UNUSED,
                         gpointer         gself)
{
  indicator_datetime_planner_emit_appointments_changed (gself);
}

static void 
on_client_view_ready (GObject      * client, 
                      GAsyncResult * res, 
                      gpointer       gself)
{
  GError * error = NULL;
  ECalClientView * view = NULL;

  if (!e_cal_client_get_view_finish (E_CAL_CLIENT (client), res, &view, &error) || error || !view)
    {
      if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
        g_warning ("indicator-datetime cannot connect to EDS client view: %s", error->message);
    }
  else
    {
      IndicatorDatetimePlannerEds * self = INDICATOR_DATETIME_PLANNER_EDS (gself);
      priv_t * p = self->priv;

      e_cal_client_view_set_flags (view, E_CAL_CLIENT_VIEW_FLAGS_NONE, NULL);
      p->views = g_slist_prepend (p->views, view);

      g_signal_connect(view, "objects-added", G_CALLBACK(on_view_objects_added), self);
      g_signal_connect(view, "objects-modified", G_CALLBACK(on_view_objects_modified), self);
      g_signal_connect(view, "objects-removed", G_CALLBACK(on_view_objects_removed), self);
      
      e_cal_client_view_start (view, &error);
      if (error)
        {
          g_warning ("Failed to start view for %s: %s\n",
				     e_source_get_uid (e_client_get_source (E_CLIENT (client))),
				     error ? error->message : "Unknown error");
		}

      /* we've got a new connected ECalClientView, so store it & notify clients */

      g_object_set_qdata_full (G_OBJECT(e_client_get_source(E_CLIENT (client))),
                               source_view_quark(),
                               view,
                               g_object_unref);

      indicator_datetime_planner_emit_appointments_changed (gself);
    }
  g_clear_error (&error);
}

static void
on_client_connected (GObject      * unused G_GNUC_UNUSED,
                     GAsyncResult * res,
                     gpointer       gself)
{
  GError * error;
  EClient * client;

  error = NULL;
  client = e_cal_client_connect_finish (res, &error);
  if (error != NULL)
    {
      if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
        g_warning ("indicator-datetime cannot connect to EDS source: %s", error->message);

      g_error_free (error);
    }
  else
    {
      /* we've got a new connected ECalClient, so store it & notify clients */

      g_object_set_qdata_full (G_OBJECT(e_client_get_source(client)),
                               source_client_quark(),
                               client,
                               g_object_unref);

      /* now create a view for it so that we can listen for changes */

      IndicatorDatetimePlannerEds * self = INDICATOR_DATETIME_PLANNER_EDS (gself);
      priv_t * p = self->priv;

      e_cal_client_get_view (E_CAL_CLIENT(client),
                             "#t", /* match all */
                             p->cancellable,
                             on_client_view_ready,
                             self);

      indicator_datetime_planner_emit_appointments_changed (gself);
    }
}

static void
on_source_enabled (ESourceRegistry * registry  G_GNUC_UNUSED,
                   ESource         * source,
                   gpointer          gself)
{
  IndicatorDatetimePlannerEds * self = INDICATOR_DATETIME_PLANNER_EDS (gself);
  priv_t * p = self->priv;

  e_cal_client_connect (source,
                        E_CAL_CLIENT_SOURCE_TYPE_EVENTS,
                        -1,
                        p->cancellable,
                        on_client_connected,
                        self);
}

static void
on_source_added (ESourceRegistry * registry,
                 ESource         * source,
                 gpointer          gself)
{
  IndicatorDatetimePlannerEds * self = INDICATOR_DATETIME_PLANNER_EDS (gself);
  priv_t * p = self->priv;

  p->sources = g_slist_prepend (p->sources, g_object_ref(source));

  if (e_source_get_enabled (source))
    on_source_enabled (registry, source, gself);
}

static void
on_source_disabled (ESourceRegistry * registry  G_GNUC_UNUSED,
                    ESource         * source,
                    gpointer          gself)
{
  ECalClient * client;
  ECalClientView * view;

  /* If this source has a connected ECalClientView, remove it & notify clients */
  if ((view = g_object_steal_qdata (G_OBJECT(source), source_view_quark())))
    {
      IndicatorDatetimePlannerEds * self = INDICATOR_DATETIME_PLANNER_EDS (gself);
      priv_t * p = self->priv;

      e_cal_client_view_stop (view, NULL);
      p->views = g_slist_remove (p->views, view);
      g_object_unref (view);
      indicator_datetime_planner_emit_appointments_changed (gself);
    }

  /* If this source has a connected ECalClient, remove it & notify clients */
  if ((client = g_object_steal_qdata (G_OBJECT(source), source_client_quark())))
    {
      g_object_unref (client);
      indicator_datetime_planner_emit_appointments_changed (gself);
    }
}

static void
on_source_removed (ESourceRegistry * registry,
                   ESource         * source,
                   gpointer          gself)
{
  IndicatorDatetimePlannerEds * self = INDICATOR_DATETIME_PLANNER_EDS (gself);
  priv_t * p = self->priv;

  on_source_disabled (registry, source, gself);

  p->sources = g_slist_remove (p->sources, source);
  g_object_unref (source);
}

static void
on_source_changed (ESourceRegistry * registry  G_GNUC_UNUSED,
                   ESource         * source    G_GNUC_UNUSED,
                   gpointer          gself)
{
  indicator_datetime_planner_emit_appointments_changed (gself);
}

static void
on_source_registry_ready (GObject      * source_object  G_GNUC_UNUSED,
                          GAsyncResult * res,
                          gpointer       gself)
{
  GError * error;
  ESourceRegistry * r;

  error = NULL;
  r = e_source_registry_new_finish (res, &error);
  if (error != NULL)
    {
      if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
        g_warning ("indicator-datetime cannot show EDS appointments: %s", error->message);

      g_error_free (error);
    }
  else
    {
      IndicatorDatetimePlannerEds * self;
      priv_t * p;
      GList * l;
      GList * sources;

      self = INDICATOR_DATETIME_PLANNER_EDS (gself);
      p = self->priv;

      g_signal_connect (r, "source-added",    G_CALLBACK(on_source_added), self);
      g_signal_connect (r, "source-removed",  G_CALLBACK(on_source_removed), self);
      g_signal_connect (r, "source-changed",  G_CALLBACK(on_source_changed), self);
      g_signal_connect (r, "source-disabled", G_CALLBACK(on_source_disabled), self);
      g_signal_connect (r, "source-enabled",  G_CALLBACK(on_source_enabled), self);

      p->source_registry = r;

      sources = e_source_registry_list_sources (r, E_SOURCE_EXTENSION_CALENDAR);
      for (l=sources; l!=NULL; l=l->next)
        on_source_added (r, l->data, self);
      g_list_free_full (sources, g_object_unref);
    }
}

/***
****  GObject virtual funcs
***/

static void
my_dispose (GObject * o)
{
  IndicatorDatetimePlannerEds * self = INDICATOR_DATETIME_PLANNER_EDS (o);
  priv_t * p = self->priv;
  GSList *l;

  if (p->cancellable != NULL)
    {
      g_cancellable_cancel (p->cancellable);
      g_clear_object (&p->cancellable);
    }

  for (l = p->views; l != NULL; l = l->next)
    {
      ECalClientView *view = E_CAL_CLIENT_VIEW (l->data);
      
      if (!view)
        continue;
      
      g_signal_handlers_disconnect_by_func (view, on_view_objects_added, self);
      g_signal_handlers_disconnect_by_func (view, on_view_objects_modified, self);
      g_signal_handlers_disconnect_by_func (view, on_view_objects_removed, self);
      e_cal_client_view_stop (view, NULL);
      g_clear_object (&view);
    }
  g_slist_free (p->views);

  if (p->source_registry != NULL)
    {
      g_signal_handlers_disconnect_by_func (p->source_registry,
                                            indicator_datetime_planner_emit_appointments_changed,
                                            self);

      g_clear_object (&self->priv->source_registry);
    }

  g_slist_free_full (p->sources, g_object_unref);

  G_OBJECT_CLASS (indicator_datetime_planner_eds_parent_class)->dispose (o);
}

/***
****  Instantiation
***/

static void
indicator_datetime_planner_eds_class_init (IndicatorDatetimePlannerEdsClass * klass)
{
  GObjectClass * object_class;
  IndicatorDatetimePlannerClass * planner_class;

  object_class = G_OBJECT_CLASS (klass);
  object_class->dispose = my_dispose;

  planner_class = INDICATOR_DATETIME_PLANNER_CLASS (klass);
  planner_class->is_configured = my_is_configured;
  planner_class->activate = my_activate;
  planner_class->activate_time = my_activate_time;
  planner_class->get_appointments = my_get_appointments;
  planner_class->activate_appointment = my_activate_appointment;
  planner_class->get_appointments_finish = my_get_appointments_finish;
}

static void
indicator_datetime_planner_eds_init (IndicatorDatetimePlannerEds * self)
{
  priv_t * p;
                                   
  p = indicator_datetime_planner_eds_get_instance_private (INDICATOR_DATETIME_PLANNER_EDS (self));

  self->priv = p;

  p->cancellable = g_cancellable_new ();

  e_source_registry_new (p->cancellable,
                         on_source_registry_ready,
                         self);
}

/***
****  Public
***/

IndicatorDatetimePlanner *
indicator_datetime_planner_eds_new (void)
{
  gpointer o = g_object_new (INDICATOR_TYPE_DATETIME_PLANNER_EDS, NULL);

  return INDICATOR_DATETIME_PLANNER (o);
}
